from datetime import datetime
import numpy as np

def title():
    return "Frequency Duration Curve"

def abstract():
    return "Gets a timesseries and determines the frequency duration curves."

def inputs():
    return [['startdate','Start date for data collection','string'],
            ['enddate','End date for data collection','string'],
            ['parameterid','Parameter ID','string'],
            ['locations','Select single location for frequency duration curve','string']]

def get_indices(array, value):
    """
    Finds the location of a value in sorted array, returns two indices.
    The indices indicate the value that is <= input value and > input value

    Parameters
    ----------
    array : sorted numpy array
    value : float, int

    Returns
    -------
    i : int, index for value <= input value
    i + 1 : int, index for value > input_value
    """
    for i in range(len(array)-1):
        if array[i] <= value < array[i+1]:
            return i, i + 1
        elif array[0] > value or array[-1] < value:
            return -999.0, -999.0


def check_month(c_year, month, last_month, row):
    """
    Support function for the get_freq function. Checks whether the new month is
    1 month later than the previous month (last_month). If not c_year is filled
    with np.nan, if month number of "month" follows last_month directly the
    value is filled in in c_year.

    Parameters
    ----------
    c_year : list, values per month for the current year
    month : int, month_number of current month (e.g. 1 = January, 2 = Feb)
    last_month : int, month_number of last month (e.g. 1 = January, 2 = Feb)
    row : pandas.DataFrame-row,
        {0: index, 1: [data_values, location_id, parameter_id]}

    Return
    ------
    c_year : list, values per month for the current year
    last_month : int, month_number of last month (e.g. 1 = January, 2 = Feb)
    """
    if month == last_month + 1:
        a = month - 1
        c_year[a] = row[1].data_values
        last_month += 1
    else:
        diff = last_month + 1 - month
        for i_month in range(diff):
            last_month += 1
            a = last_month - 1
            c_year[a] = np.nan
        a = month - 1
        c_year[a] = row[1].data_values
        last_month += 1
    return c_year, last_month


def get_freq(df):
    """
    Determines input for frequency and duration plots
    The following method is used for calculation:
    1. get data
    2. sort data in ascending order
    3. determine rank and calculate 1/(meff + 1) (assuming equiprobability)
    4. Interpolate to 0.1, 0.2, 0.3 etc. (frequency plot)

    5. sort results from 4 to ascending order (duration plot)

    Parameters
    ----------
    df : pandas.DataFrame, rows are times, column with values should be called
        data_values

    Returns
    -------
    perc_list : list, percentages to be plotted
    df_freq : pandas.DataFrame, frequency data (step 4)
    df_dur : pandas.DataFrame, duration data (step 5)
    """
    df = df.sort_index()
    data = []
    years = [df.index[0].year]
    last_year = df.index[0].year
    last_month = 0
    c_year = np.zeros(12)
    for row in df.iterrows():
        year = row[0].year
        month = row[0].month
        if year == last_year:
            c_year, last_month = check_month(c_year, month, last_month, row)
        else:
            years.append(year)
            data.append(c_year)
            last_year = np.copy(year)
            c_year = np.zeros(12)
            last_month = 0
            c_year, last_month = check_month(c_year, month, last_month, row)

    X = np.array(data)
    R = np.sort(X, axis=0)[::-1]
    m_eff = np.array([np.count_nonzero(~np.isnan(R[:, j])) for j in range(12)])
    F = np.zeros((R.shape))
    for i in xrange(len(R)):
        new_row = (i + 1.) / (m_eff + 1)
        F[i, :] = new_row
    shape_y, shape_x = F.shape
    perc_list = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    perc_list_str = [str(perc) for perc in perc_list]
    B = np.zeros((len(perc_list), shape_x))
    for i, perc in enumerate(perc_list):
        for col in xrange(shape_x):
            column = F[:, col]
            ind1, ind2 = get_indices(column, perc)
            factor = ((perc - F[ind1, col]) / (F[ind2, col] - F[ind1, col]))
            b = (R[ind2, col] - R[ind1, col])
            B[i, col] = R[ind1, col] + b * factor

    month_names = ["Jan", "Feb", "Mar", "Apr", "May", "Jun",
                   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]
#    df_m = pd.DataFrame(month_names, index=month_names, columns=["months"])
    df_m = pd.DataFrame(range(1, 13), index=month_names, columns=["months"])
    df = pd.DataFrame(data=np.transpose(B), index=month_names,
                      columns=perc_list_str[::-1])
    df_freq = pd.concat([df_m, df], axis=1)

    B_sort = np.sort(B)
    df2 = pd.DataFrame(data=np.transpose(B_sort), index=month_names,
                       columns=perc_list_str[::-1])
    df_dur = pd.concat([df_m, df2], axis=1)

    return perc_list, df_freq, df_dur


def bokeh_frequency_duration(parameter, locations, startdate, enddate,
                             mmode="year", nmode="month", wps=True,
                             mode='static', static_folder="", js_folder=""):
    """
    Creates frequency and duration plots based on a selected parameter and
    location.
    The following method is used for calculation:
    1. get data
    2. sort data in ascending order
    3. determine rank and calculate 1/(meff + 1) (assuming equiprobability)
    4. Interpolate to 0.1, 0.2, 0.3 etc. (frequency plot)

    5. sort results from 4 to ascending order (duration plot)

    Parameters
    ----------
    parameter : string, name of parameter that is selected
    locations : list, list of locations (strings) that are selected in the
        viewer
    startdate : start date of search period
    enddate : end date of search period
    mmode : string, default "year"
        time-unit of main axis
    nmode : string, default "month"
        time-unit of secondary axis
    wps : boolean, default True
    mode : string, default static
        determines output of script (static, wps)
    static_folder : string, default .
        Path to the static folder on server

    Returns
    -------
    f : StringIO object containing json with path to html of bokeh plot
    """

    sources_dict = {}
    for key, grp in df.groupby(df.location_id):
        perc_list, df_freq, df_dur = get_freq(grp)
        sources_dict[key] = {}
        sources_dict[key]["freq"] = df_freq
        sources_dict[key]["dur"] = df_dur

#    return X, R, F, B, perc_list, df_new, sources_dict
    # plot

    i = 0
    perc_list_str = [str(perc) for perc in perc_list]

    for perc in perc_list_str:
        p_freq.line("months", perc, source=source["freq"], color=colors[i], legend=perc)
        p_dur.line("months", perc, source=source["dur"], color=colors[i], legend=perc)
        i += 1
    param = param_dict[parameter].name
    p_freq.yaxis.axis_label = "%s" % param
    p_dur.yaxis.axis_label = "%s" % param
    p_dur.legend.location = "top_left"

    grd = bokeh.plotting.gridplot([[p_freq, p_dur]])

    f = generate_output(static_folder, js_folder, mode, plot_name, grd)

    return f

#def process(identifier, parameters, records):
#    values = []
#    for record in records:
#        values.append(record['value'])
#    mean = np.mean(values)
#    for record in records:
#        record['date'] = datetime.strftime(record['date'], "%Y-%m-%d %H:%M:%S") # change datetime.datetime to str for json.dumps()
#        record['value'] = record['value'] - mean
#        
#    return records
